{% if web_apps %}
    {% for app in web_apps %}
        {% if app.default %}
c.ServerApp.default_url = '{{ app.path }}'
        {% endif %}
    {% endfor %}

c.ServerProxy.servers = {
    {% for app in web_apps %}
    '{{ app.path }}': {
        'command': {{ app.cmdline }},
        {% if app.port %}
        'port': app.port,
        {% endif %}
        {% if app.timeout %}
        'timeout': {{ app.timeout }},
        {% endif %}
        {% if app.absolute_url %}
        'absolute_url': True,
        {% endif %}
    },
    {% endfor %}
}
{% endif %}



"""Generators generate the final Dockerfile based on the settings.

The following generators are currently supported:

* :mod:`~ou_container_builder.generators.jupyter_notebook`
* :mod:`~ou_container_builder.generators.web_app`
"""
from . import jupyter_notebook  # noqa
from . import web_app  # noqa



"""Configuration validator for the OU Container Builder ContainerConfig.yaml."""
import shlex

from cerberus import Validator
from typing import Union


def multiline_splitter(value: str):
    """Split the value on new-lines.

    :param value: The string value to split
    :type value: str
    :return: The split list
    :rtype: list
    """
    return value.split('\n')


def ensure_list(splitter):
    """Create a coercion function that ensures the value is a list.

    :param splitter: The splitting function to apply to strings
    :type splitter: callable
    :return: The split list or the original value
    :rtype: list
    """
    def coercer(value):
        """Coerce the value into list format."""
        if isinstance(value, str):
            return splitter(value)
        elif isinstance(value, list):
            return value
        else:
            return [value]

    return coercer


def filter_empty_items(value):
    """Filter empty values from the list."""
    return [line for line in value if line.strip()]


schema = {
    'module': {
        'type': 'dict',
        'required': True,
        'schema': {
            'code': {
                'type': 'string',
                'required': True
            },
            'presentation': {
                'type': 'string',
                'required': True
            }
        }
    },
    'type': {
        'type': 'string',
        'required': True,
        'allowed': ['jupyter-notebook', 'web-app']
    },
    'base': {
        'type': 'string',
        'required': False,
        'default': '3.9'
    },
    'content': {
        'type': 'list',
        'schema': {
            'type': 'dict',
            'required': True,
            'schema': {
                'source': {
                    'type': 'string',
                    'required': True,
                    'empty': False
                },
                'target': {
                    'type': 'string',
                    'default': ''
                },
                'overwrite': {
                    'type': 'string',
                    'required': True,
                    'allowed': ['always', 'never', 'if-unchanged']
                }
            }
        }
    },
    'sources': {
        'type': 'dict',
        'schema': {
            'apt': {
                'type': 'list',
                'schema': {
                    'type': 'dict',
                    'schema': {
                        'name': {
                            'type': 'string',
                            'required': True,
                            'empty': False
                        },
                        'key': {
                            'type': 'string',
                            'required': True,
                            'empty': False,
                        },
                        'deb': {
                            'type': 'string',
                            'required': True,
                            'empty': False,
                        }
                    }
                }
            }
        }
    },
    'packages': {
        'type': 'dict',
        'schema': {
            'apt': {
                'type': 'list',
                'schema': {
                    'type': 'string'
                }
            },
            'pip': {
                'type': 'list',
                'schema': {
                    'type': 'string'
                }
            }
        }
    },
    'scripts': {
        'type': 'dict',
        'schema': {
            'build': {
                'type': 'list',
                'schema': {
                    'type': 'dict',
                    'schema': {
                        'commands': {
                            'type': 'list',
                            'schema': {
                                'type': 'string'
                            },
                            'coerce': (ensure_list(multiline_splitter), filter_empty_items)
                        }
                    }
                }
            },
            'startup': {
                'type': 'list',
                'schema': {
                    'type': 'dict',
                    'schema': {
                        'commands': {
                            'type': 'list',
                            'schema': {
                                'type': 'string'
                            },
                            'coerce': (ensure_list(multiline_splitter), filter_empty_items)
                        }
                    }
                }
            },
            'shutdown': {
                'type': 'list',
                'schema': {
                    'type': 'dict',
                    'schema': {
                        'commands': {
                            'type': 'list',
                            'schema': {
                                'type': 'string'
                            },
                            'coerce': (ensure_list(multiline_splitter), filter_empty_items)
                        }
                    }
                }
            }
        }
    },
    'jupyter_notebook': {
        'type': 'dict',
        'schema': {
            'default_url': {
                'type': 'string',
                'empty': False
            }
        }
    },
    'web_apps': {
        'type': 'list',
        'schema': {
            'type': 'dict',
            'schema': {
                'path': {
                    'type': 'string',
                    'required': True,
                    'empty': False
                },
                'cmdline': {
                    'type': 'list',
                    'required': True,
                    'empty': False,
                    'schema': {
                        'type': 'string',
                        'required': True,
                        'empty': False
                    },
                    'coerce': (ensure_list(shlex.split), filter_empty_items)
                },
                'port': {
                    'type': 'integer',
                    'default': 0
                },
                'default': {
                    'type': 'boolean',
                    'default': False
                },
                'timeout': {
                    'type': 'integer'
                },
                'absolute_url': {
                    'type': 'boolean',
                    'default': False
                }
            }
        }
    },
    'services': {
        'type': 'list',
        'schema': {
            'type': 'string',
            'empty': False
        }
    },
    'packs': {
        'type': 'list',
        'schema': {
            'type': 'string',
            'allowed': ['tutorial-server', 'mariadb']
        }
    },
    'tutorial_server': {
        'type': 'dict',
        'schema': {
            'parts': {
                'type': 'list',
                'schema': {
                    'type': 'dict',
                    'schema': {
                        'name': {
                            'type': 'string',
                            'required': True,
                            'empty': False,
                        },
                        'type': {
                            'type': 'string',
                            'required': True,
                            'allowed': ['tutorial', 'workspace', 'live']
                        },
                        'path': {
                            'type': 'string',
                            'required': True
                        }
                    }
                }
            },
            'default': {
                'type': 'string',
                'required': True,
                'empty': False
            },
            'php-cgi': {
                'type': 'boolean',
                'required': False,
                'default': False
            }
        }
    },
    'mariadb': {
        'type': 'dict',
        'schema': {
            'database': {
                'type': 'string',
                'required': True,
                'empty': False,
            },
            'username': {
                'type': 'string',
                'required': True,
                'empty': False,
            },
            'password': {
                'type': 'string',
                'required': True,
                'empty': False,
            }
        }
    },
    'hacks': {
        'type': 'list',
        'schema': {
            'type': 'string',
            'required': True,
            'empty': False,
            'allowed': ['missing-man1']
        }
    },
    'flags': {
        'type': 'dict',
        'schema': {
            'ou_container_content': {
                'type': 'boolean',
                'default': False
            }
        }
    }
}


def validate_settings(settings: dict) -> Union[dict, bool]:
    """Validate the configuration settings against the configuration schema.

    :param settings: The settings parsed from the configuration file
    :type settings: dict
    :return: The validated and normalised settings if they are valid, otherwise ``False``
    :rtype: boolean or dict
    """
    validator = Validator(schema)
    if settings and validator.validate(settings):
        return validator.document
    elif settings is None:
        return ['Your configuration file is empty']
    else:
        error_list = []

        def walk_error_tree(err, path):
            if isinstance(err, dict):
                for key, value in err.items():
                    walk_error_tree(value, path + (str(key), ))
            elif isinstance(err, list):
                for sub_err in err:
                    walk_error_tree(sub_err, path)
            else:
                error_list.append(f'{".".join(path)}: {err}')

        walk_error_tree(validator.errors, ())
        return error_list



"""Build containers based on a Jupyter Notebook."""
import os

from jinja2 import Environment

from ou_container_builder.utils import merge_settings


def setup(context: str, env: Environment, settings: dict) -> dict:
    """Run the setup for the Jupyter Notebook generator.

    Merges the required settings for Jupyter Notebooks into the settings.

    :param context: The context path within which the generation is running
    :type context: str
    :param env: The Jinja2 environment to use for loading and rendering templates
    :type env: :class:`~jinja2.environment.Environment`
    :param settings: The settings parsed from the configuration file
    :type settings: dict
    :return: The updated settings
    :rtype: dict
    """
    settings = merge_settings(settings, {
        'packages': {
            'pip': [
                'jupyterhub==1.3.0',
                'notebook>=6.0.0,<7'
            ]
        },
        'content': [
            {
                'source': 'ou-builder-build/start-notebook.sh',
                'target': '/usr/bin/start.sh',
                'overwrite': 'always'
            },
            {
                'source': 'ou-builder-build/jupyter_notebook_config.py',
                'target': '/etc/jupyter/jupyter_notebook_config.py',
                'overwrite': 'always'
            }
        ],
        'scripts': {
            'build': [
                {
                    'commands': [
                        'chmod a+x /usr/bin/start.sh'
                    ]
                }
            ]
        }
    })
    return settings


def generate(context: str, env: Environment, settings: dict):
    """Generate the Dockerfile for a Jupyter Notebook container.

    This generates the ``Dockerfile``, ``jupyter_notebook_config.py``, and ``start-notebook.sh`` files.

    :param context: The context path within which the generation is running
    :type context: str
    :param env: The Jinja2 environment to use for loading and rendering templates
    :type env: :class:`~jinja2.environment.Environment`
    :param settings: The settings parsed from the configuration file
    :type settings: dict
    """
    with open(os.path.join(context, 'ou-builder-build', 'jupyter_notebook_config.py'), 'w') as out_f:
        tmpl = env.get_template('generators/jupyter_notebook/jupyter_notebook_config.py')
        out_f.write(tmpl.render(**settings))

    with open(os.path.join(context, 'ou-builder-build', 'start-notebook.sh'), 'w') as out_f:
        tmpl = env.get_template('generators/jupyter_notebook/start.sh')
        out_f.write(tmpl.render(**settings))

    with open(os.path.join(context, 'Dockerfile'), 'w') as out_f:
        tmpl = env.get_template('Dockerfile.jinja2')
        out_f.write(tmpl.render(**settings))



c.NotebookApp.ip = '0.0.0.0'
c.NotebookApp.notebook_dir = '/home/ou/{{ module.code }}-{{ module.presentation }}'
c.NotebookApp.open_browser = False
c.NotebookApp.quit_button = False
c.JupyterHub.shutdown_on_logout
c.NotebookApp.iopub_data_rate_limit = 10000000
c.MappingKernelManager.cull_idle_timeout = 1200
c.MappingKernelManager.cull_interval = 10
{% if jupyter_notebook and 'default_url' in jupyter_notebook %}
c.NotebookApp.default_url = '{{ jupyter_notebook['default_url'] }}'
{% endif %}

{% if web_apps %}
c.ServerProxy.servers = {
    {% for app in web_apps %}
    '{{ app.path }}': {
        'command': {{ app.cmdline }},
        {% if app.port %}
        'port': app.port,
        {% endif %}
        {% if app.timeout %}
        'timeout': {{ app.timeout }},
        {% endif %}
        {% if app.absolute_url %}
        'absolute_url': True,
        {% endif %}
    },
    {% endfor %}
}
{% endif %}



"""Packs group together related settings and configuration files.

Packs are split into two group. The core packs :mod:`~ou_container_content.packs.services` and
:mod:`~ou_container_content.packs.content`, which are automatically included and the optional
packs :mod:`~ou_container_content.packs.tutorial_server` and :mod:`~ou_container_content.packs.mariadb`,
which must be explicitly included in the ``ContainerConfig.yaml``.
"""
from .tutorial_server import apply_pack as tutorial_server  # noqa
from .mariadb import apply_pack as mariadb  # noqa
from .services import apply_pack as services  # noqa
from .content import apply_pack as content  # noqa
from .scripts import apply_pack as scripts  # noqa



"""Pack to handle setting up content in the container."""
import os

from jinja2 import Environment

from ..utils import merge_settings


def apply_pack(context: str, env: Environment, settings: dict) -> dict:
    """Apply the content pack.

    This ensures that the OU Container Content application is integrated into the container and that all configured
    files are distributed in the container as configured. This also sets the ``ou_container_content`` flag to ensure
    that the OU Container Content application is run upon container startup.

    :param context: The context path within which the generation is running
    :type context: str
    :param env: The Jinja2 environment to use for loading and rendering templates
    :type env: :class:`~jinja2.environment.Environment`
    :param settings: The settings parsed from the configuration file
    :type settings: dict
    :return: The updated settings
    :rtype: dict
    """
    if 'content' in settings and settings['content']:
        settings = merge_settings(settings, {
            'flags': {
                'ou_container_content': True
            }
        })
    if settings['flags'] and settings['flags']['ou_container_content']:
        settings = merge_settings(settings, {
            'packages': {
                'pip': [
                    'ou-container-content>=1.0.4'
                ]
            },
            'scripts': {
                'build': [
                    {
                        'commands': '''ou-container-content prepare'''
                    }
                ]
            }
        })
        with open(os.path.join(context, 'ou-builder-build', 'content_config.yaml'), 'w') as out_f:
            tmpl = env.get_template('content_config.yaml')
            out_f.write(tmpl.render(**settings))
    return settings



"""Build containers housing one or more web applications."""
import os

from jinja2 import Environment

from ..utils import merge_settings


def setup(context: str, env: Environment, settings: dict) -> dict:
    """Run the setup for the Web Applications generator.

    Merges the required settings for Web Applications into the settings.

    :param context: The context path within which the generation is running
    :type context: str
    :param env: The Jinja2 environment to use for loading and rendering templates
    :type env: :class:`~jinja2.environment.Environment`
    :param settings: The settings parsed from the configuration file
    :type settings: dict
    :return: The updated settings
    :rtype: dict
    """
    settings = merge_settings(settings, {
        'packages': {
            'pip': [
                'jupyter-server-proxy',
                'jupyterhub==1.3.0'
            ]
        },
        'content': [
            {
                'source': 'ou-builder-build/start-web-app.sh',
                'target': '/usr/bin/start.sh',
                'overwrite': 'always'
            },
            {
                'source': 'ou-builder-build/jupyter_server_config.py',
                'target': '/etc/jupyter/jupyter_server_config.py',
                'overwrite': 'always'
            }
        ],
        'scripts': {
            'build': [
                {
                    'commands': [
                        'chmod a+x /usr/bin/start.sh'
                    ]
                }
            ]
        }
    })
    return settings


def generate(context: str, env: Environment, settings: dict):
    """Generate the Dockerfile for one or more web applications.

    This generates the ``Dockerfile``, ``start-web-app.sh``, and ``jupyter_server_config.py`` files.

    :param context: The context path within which the generation is running
    :type context: str
    :param env: The Jinja2 environment to use for loading and rendering templates
    :type env: :class:`~jinja2.environment.Environment`
    :param settings: The settings parsed from the configuration file
    :type settings: dict
    """
    with open(os.path.join(context, 'ou-builder-build', 'start-web-app.sh'), 'w') as out_f:
        tmpl = env.get_template('generators/web_app/start.sh')
        out_f.write(tmpl.render(**settings))

    with open(os.path.join(context, 'ou-builder-build', 'jupyter_server_config.py'), 'w') as out_f:
        tmpl = env.get_template('generators/web_app/jupyter_server_config.py')
        out_f.write(tmpl.render(**settings))

    with open(os.path.join(context, 'Dockerfile'), 'w') as out_f:
        if settings['type'] == 'web-app':
            tmpl = env.get_template('Dockerfile.jinja2')
        out_f.write(tmpl.render(**settings))



"""The main OU Container Builder commandline application.

Run ``ou-container-builder --help`` for help with the command-line parameters.
"""
import click
import os
import re
import shutil
import subprocess

from jinja2 import Environment, PackageLoader
from yaml import load
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from . import generators, packs
from .validator import validate_settings


def run_build(settings: dict, context: str, build: bool, clean: bool, tag: list) -> list:
    """Run the build process.

    This processes the ``settings``, generates the required files, and then runs the Docker build process.

    :param settings: The settings parsed from the configuration file
    :type settings: dict
    :param context: The directory within which to run the build
    :type context: str
    :param build: Whether to automatically invoke ``docker build``
    :type build: bool
    :param clean: Whether to automatically clean all generated files
    :type clean: bool
    :param tag: A list of tags to pass to docker
    :type tag: list[str]
    :return: A list with any errors that occured during processing
    :rtype: list
    """
    settings = validate_settings(settings)
    if isinstance(settings, dict):
        env = Environment(loader=PackageLoader('ou_container_builder', 'templates'),
                          autoescape=False)

        if os.path.exists(os.path.join(context, 'ou-builder-build')):
            shutil.rmtree(os.path.join(context, 'ou-builder-build'))
        os.makedirs(os.path.join(context, 'ou-builder-build'))

        # Handle base image
        if re.match(r'3\.[789]', settings['base']):
            settings['base'] = f'python:{settings["base"]}-slim-buster'

        # Handle optional packs
        if 'packs' in settings and settings['packs']:
            if 'tutorial-server' in settings['packs']:
                settings = packs.tutorial_server(context, env, settings)
            if 'mariadb' in settings['packs']:
                settings = packs.mariadb(context, env, settings)

        # Setup the generators
        if settings['type'] == 'jupyter-notebook':
            settings = generators.jupyter_notebook.setup(context, env, settings)
        elif settings['type'] == 'web-app':
            settings = generators.web_app.setup(context, env, settings)

        # Handle core packs
        if 'services' in settings:
            settings = packs.services(context, env, settings)
        if 'scripts' in settings:
            settings = packs.scripts(context, env, settings)
        settings = packs.content(context, env, settings)

        # Handle automatic hacks
        if 'packages' in settings and 'apt' in settings['packages']:
            if 'openjdk-11-jdk' in settings['packages']['apt']:
                if 'hacks' in settings:
                    if 'missing-man1' not in settings['hacks']:
                        settings['hacks'].append('missing-man1')
                else:
                    settings['hacks'] = ['missing-man1']

        settings = validate_settings(settings)
        if isinstance(settings, dict):
            # Sort package lists
            if 'packages' in settings:
                if 'apt' in settings['packages']:
                    settings['packages']['apt'].sort()
                if 'pip' in settings['packages']:
                    settings['packages']['pip'].sort()

            # Generate the Dockerfiles
            if settings['type'] == 'jupyter-notebook':
                generators.jupyter_notebook.generate(context, env, settings)
            elif settings['type'] == 'web-app':
                generators.web_app.generate(context, env, settings)

            if build:
                cmd = ['docker', 'build', context]
                if tag:
                    for t in tag:
                        cmd.append('--tag')
                        cmd.append(t)
                subprocess.run(cmd)
                if clean:
                    os.unlink(os.path.join(context, 'Dockerfile'))
                    if os.path.exists(os.path.join(context, 'ou-builder-build')):
                        shutil.rmtree(os.path.join(context, 'ou-builder-build'))
            return []
        else:
            return settings
    else:
        return settings


@click.command()
@click.option('-c', '--context',
              default='.',
              help='Context within which the container will be built',
              show_default=True)
@click.option('-b/-nb', '--build/--no-build',
              default=True,
              help='Automatically build the container',
              show_default=True)
@click.option('--clean/--no-clean',
              default=True,
              help='Automatically clean up after building the container',
              show_default=True)
@click.option('--tag',
              multiple=True,
              help='Automatically tag the generated image')
def main(context: str, build: bool, clean: bool, tag: list):
    """Build your OU Container."""
    with open(os.path.join(context, 'ContainerConfig.yaml')) as config_f:
        settings = load(config_f, Loader=Loader)
    result = run_build(settings, context, build, clean, tag)
    if result:
        click.echo(click.style('There are errors in your configuration settings:', fg='red'), err=True)
        click.echo(err=True)
        for error in settings:
            click.echo(error, err=True)


if __name__ == '__main__':
    main()
